/*
 *
 * autonice.c for OSF/1 and /procfs
 *
 * (C) R.P.Bellis 1993
 * 
 */

#define _OSF_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/dir.h>

#ifdef OSF
#include <sys/limits.h>
#endif /* OSF */
#ifdef LINUX
#include <limits.h>
#endif /* LINUX */

#include <sys/ioctl.h>
#include <sys/procfs.h>
#include <sys/times.h>
#include <sys/file.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>
#include <math.h>
#include <pwd.h>
#include <ctype.h>

#ifdef LINUX
#include <sys/stat.h>
#include <dirent.h>
#include "linux.h"

#define TRUE 1
#define FALSE 0
#define FNAME_SZ 256
#endif /* LINUX */

#define MAX_CONFIG 200 /* How many lines  you can put in the config. file */
#define DEFAULT_TIME 30.0  /* Default convergence time if not specified on
			      command line */

/*@unused@*/
static char RCSid[] = "$Id: autonice.c,v 1.7 2004/05/01 07:33:23 dpinson Exp $";

struct EntryStruct {
  char UserName[10];
  char ProcessName[FNAME_SZ];
  int  MinPriority;
  int  MaxPriority;
};

struct  ConfigEntry{
  int NumEntries;
  struct EntryStruct EntryArray[MAX_CONFIG];
} Config;

/* Definition of function to fork daemon */
extern int daemon(int nochdir, int noclose);

/* ARGSUSED */
static void terminate(int status)
{
	syslog(LOG_INFO, "terminated");
	exit(status);
}

static int ReadConfig( struct ConfigEntry *Config)
{

  char *FileName;
  char InputLine[100];
  char OutputMessage[100];
  int  Result;
  int  Index = 0;
  int  Line = 0;

  FILE *ConfigFile;

  FileName = getenv ("AUTONICE_CONFIG");

  Config->NumEntries = 0;

  if (!FileName) {
    syslog(LOG_INFO,"no config file defined in AUTONICE_CONFIG environment variable; reniceing all the processes");
    return 0;
  }

  snprintf(OutputMessage,99,"Using config file %s",FileName);
  syslog(LOG_INFO,OutputMessage);

  if ((ConfigFile = fopen(FileName,"r")) == NULL) {
    snprintf(OutputMessage,99,"unable to open %s config file",FileName);
    syslog(LOG_INFO, OutputMessage);
    return 0;
  }

  /* successful open of config file */
  do {
    (void) fgets(InputLine, (int) sizeof(InputLine), ConfigFile);
    Line++;

    if ( (sizeof(InputLine) > 2) & (InputLine[0] != '#')) {
      /* ok, not an empty or comment line */
/*
 * why is the zeroing of this portion of memory required? Under Linux
 * this will result in the fopen(proc) exiting with an out of memory 
 * error. I need to trace that a bit better. If someone has a better 
 * idea than me of what is going on I would appreciate it.
 */

#ifdef OSF
      memset(&(Config->EntryArray[Index]),0,sizeof(struct ConfigEntry));
#endif /* OSF */

      Result = sscanf(InputLine,"%10s %256s %d %d",
		      Config->EntryArray[Index].UserName,
		      Config->EntryArray[Index].ProcessName,
		      &(Config->EntryArray[Index].MinPriority),
		      &(Config->EntryArray[Index].MaxPriority) );
      if (Result != 4) {
	snprintf(OutputMessage,99,"error parsing line %d (%s) in config file ",
		Line,InputLine);
        syslog(LOG_INFO,OutputMessage);
	syslog(LOG_INFO,"Discarding all the entries in the file ");
        fclose (ConfigFile);
        return 0;
      }
      if (Index < MAX_CONFIG -1 )
	{
	  Index++;
	  Config->NumEntries++;
	}
      else {
        snprintf(OutputMessage,99,"maximum (%d) config lines read", MAX_CONFIG);
        syslog(LOG_INFO, OutputMessage);
	syslog(LOG_INFO,"Discarding all the entries in the file ");
        fclose (ConfigFile);
        return 0;
      }
    }
  } while (feof(ConfigFile) != 0);
  
  fclose (ConfigFile);
  return 1; 
}

#if defined( OSF )
int CheckConfig(struct ConfigEntry *Config,
		int *np, int p, prpsinfo_t *ps,
		struct passwd *Passwd)
#elif defined( LINUX )
static int CheckConfig(struct ConfigEntry *Config,
		int *np, int p, struct linprocdata *ps,
		struct passwd *Passwd)
#endif
{
  int Index;
  int UserOK;
  int ProcessOK;
/*  char OutputMessage[100]; not being used Linux thing..remove ? */

  for (Index = 0; Index < Config->NumEntries; Index++)
    {
      if (strcmp(Config->EntryArray[Index].UserName,Passwd->pw_name) != 0 ||
	  strcmp(Config->EntryArray[Index].UserName,"*") != 0)  
	  UserOK = TRUE;
      else 
	  UserOK=FALSE;
    
      if (strcmp(Config->EntryArray[Index].ProcessName,ps->pr_fname) != 0 ||
	  strcmp(Config->EntryArray[Index].ProcessName,"*") != 0) 
	  ProcessOK = TRUE;
      else 
	  ProcessOK = FALSE;

      if ( UserOK & ProcessOK ) {

	if ( (*np <= Config->EntryArray[Index].MinPriority) &&
	   (*np >= Config->EntryArray[Index].MaxPriority) )
	  return 1; /* ok, set priority to np on exit */
	else
	  {
	    if (*np < Config->EntryArray[Index].MaxPriority)
	      {
		*np = Config->EntryArray[Index].MaxPriority;
		return 1;
	      }
	    if (*np > Config->EntryArray[Index].MinPriority)
	      {
		*np = Config->EntryArray[Index].MinPriority;
		return 1;
	      }
	    return 0; /* don't set */
	  }
      }
    }
  return 0;   /* don't set priority to np on exit */
      
}


/* ARGSUSED */
int main(int argc, char *argv[])
{
	struct utsname	 uts;
	int		 osf1;
	size_t           Index;
	float		 osf1_release = 0.0;
	double           Convergence, ConvTime, ConvFactor;
	int		 millisec_time;
	DIR		 *dirp;
	int              ConfigEnabled, CheckOK, NumError;
	char             OutputMessage[100];
	char             Number[20];
        struct passwd    *Passwd;

	/* Read system type */
	if (uname(&uts) == -1) {
		perror("uname");
		return EXIT_FAILURE;
	}

	/* Detect OSF/1 v3.0 */
	osf1 = !strcmp(uts.sysname, "OSF1");
	if (osf1) {
		osf1_release = atof(&uts.release[1]);
	}
	millisec_time = ((osf1_release >= 3.0) && (osf1_release < 4.0) );

	/* Fork as a daemon */
	if (daemon(0, 0) == -1) {
		perror("fork");
		return EXIT_FAILURE;
	}

#ifdef OSF
	/* Open log file */
	if (openlog("autonice", LOG_PID, LOG_DAEMON) == -1) {
		perror("openlog");
		return EXIT_FAILURE;
	}
#endif /* OSF */
#ifdef LINUX
        /* note: openlog should return void ?(osf as well as linux) */
        /* gcc complains dec cc doesn't...this is ok for the moment  */
	openlog("autonice", LOG_PID, LOG_DAEMON);
#endif /* LINUX */

	/* Log daemon start up */
	syslog(LOG_INFO, "started");

	/* Install abnormal signal handlers */
	(void)signal(SIGHUP, terminate);
	(void)signal(SIGTERM, terminate);

	/* Go to the /proc file system */
	if (chdir("/proc") != 0) {
		syslog(LOG_INFO, "chdir: %m");
		terminate(EXIT_FAILURE);
	}

	NumError = FALSE;

        ConfigEnabled = ReadConfig ( &Config );

	if (argc == 2) {
	  /* let's see if it is a convergence factor */
	  strncpy (Number, argv[1], 20);
	  Number[sizeof(Number)-1] = 0;
	  for (Index = 0; Index < strlen(Number); Index++) {
	    if (!isdigit(Number[Index])) {
	      fprintf(stderr,"Not a number\n");
	      /* well, not a number, use standard convergence time */
	      ConvTime = DEFAULT_TIME ;  /* 30 minutes */
	      NumError = TRUE;
              break;
	    }
	  }
	  if (NumError == TRUE) {
	    ConvTime = DEFAULT_TIME ;
	    snprintf(OutputMessage, 99,
                    "Error, the convergence time %s is not a number, using standard value (%4.0f)", 
                    Number, ConvTime);
	    syslog(LOG_INFO, OutputMessage);
	  } else {
	    ConvTime = atof(Number);
	    snprintf(OutputMessage, 99,"Ok, the convergence time will be in %4.0f minutes", ConvTime);
	    syslog(LOG_INFO, OutputMessage);
	  	  }
	}
	else {
	  ConvTime = 30.0  ;
	  syslog(LOG_INFO,"Using standard convergence time (%4.0f minutes)",
		 ConvTime);
	}
	Convergence = exp( log (ConvTime) / 5.0);
	ConvFactor = 4.0 / log (Convergence);
		      
	/* Open the directory file */
#ifdef OSF
	if (!(dirp = opendir("."))) {
		syslog(LOG_INFO, "opendir: %m");
		terminate(EXIT_FAILURE);
	}
#endif /* OSF */
#ifdef LINUX
	if ((dirp = opendir(".")) == NULL) {
		syslog(LOG_INFO, "opendir: %m");
		terminate(EXIT_FAILURE);
	}
#endif /* LINUX */

	while (1) {
		struct dirent	*dp;
		while ((dp = readdir(dirp))!=NULL) {
#ifdef OSF
			int		 fd;
			prpsinfo_t	 ps;
#endif /* OSF */
#ifdef LINUX
        		/* use the linux struct to hold process data */
                        struct linprocdata ps;
#endif /* LINUX */
			double		 cputime;
			int		 p, np;
			char		*path = dp->d_name;

			/* Check that it's a process file */
			if (!isdigit(path[0])) continue;

			/* NOTE: DEC OSF/1 proc(4) claims write access
			   is neccesary for ioctl(2) on /proc entries.
			   This appears not to be the case, and opening
			   with O_RDWR invokes a serious memory leak in
			   the kernel under V3.0 */

			/* Check that the file exists and is accessible */
			if (access(path, R_OK) < 0) continue;
#ifdef OSF
			/* Open the file */
			if ((fd = open(path, O_RDONLY)) < 0) continue;

			/* Read the process info */
			if (ioctl(fd, PIOCPSINFO, &ps) < 0) continue;

			/* Close the file */
			close(fd);
#endif /* OSF */
#ifdef LINUX
                        /*on linux, call linuxstuff */
                        if (linuxproc(path,&ps) < 0) continue;
#endif /* LINUX */			
			/* Don't do anything to root's processes */
			if (ps.pr_uid == 0) continue;

			/* Calculate CPU time in seconds */
			cputime = (double)ps.pr_time.tv_sec +
                                  (double)ps.pr_time.tv_nsec / 1.0E9;

			/* NOTE: Under DEC OSF/1 V3.x the pr_time field
			   appears to be the time in milliseconds rather
			   than seconds */
			if (millisec_time) {
				cputime /= 1000.0;
			}

			/* Calculate automatic priority level */
			if (cputime > 60.0) {
				np =  (int)(log(cputime/60.0) * ConvFactor);
				if (np < 0) np = 0;
				if (np > 19) np = 19;
			} else {
				np = 0; continue;
			}

			/* Get the current priority level */
			p = getpriority(PRIO_PROCESS, ps.pr_pid);

			/* Is the priority is already correct ? */
			if (p >= np)  continue;

			Passwd = getpwuid(ps.pr_uid);

			CheckOK = CheckConfig(&Config,&np,p,&ps, Passwd);
			if (p >= np)  continue;

			if ( ConfigEnabled && !CheckOK) continue;

			/* Actually change the priority and report on success */

			if (setpriority(PRIO_PROCESS, ps.pr_pid, np) == -1) {
				syslog(LOG_INFO,
				       "error setting priority for %s[%d]",
					ps.pr_fname, ps.pr_pid);
			} else {
			  /* aggiungere un'altra variabile "np"in input a
			     checkconfig in modo da poter avere anche un log sui
			     valori presi dal file di configurazione */
				syslog(LOG_INFO,
					"setting priority for %s[Pid=%d User:%s Cputime=%6.0f minutes] to %d",
					ps.pr_fname, ps.pr_pid,
				       Passwd->pw_name, cputime/60.0,np);
			} 
#ifdef LINUX
                       /* free strings in the linux struct */
                        linuxprocfree(&ps);
#endif /* LINUX */
		}

		rewinddir(dirp);
		sleep(60);
	}

	closedir(dirp);
	closelog();

	terminate(EXIT_FAILURE);
}
